/*
 *  EchoCS by Davide Libenzi ( Simple TCP/IP Echo Client && Server )
 *  Copyright (C) 2003	Davide Libenzi
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Davide Libenzi <davidel@xmailserver.org>
 *
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/signal.h>
#include <sys/ioctl.h>
#include <sys/resource.h>
#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <arpa/nameser.h>
#include <netdb.h>
#include <errno.h>
#include <event.h>
#include "dbllist.h"



#define ECS_STD_SERVER_PORT 8080
#define ECS_STD_LISTEN_SIZE 1024
#define ECS_STD_CLIENT_CONNS 1
#define ECS_SAMPLE_TIME 1000000ULL
#define ECS_STD_REQ_SIZE 128
#define ECS_MAX_REQ_SIZE 1024



enum ecs_states {
	ecss_start,
	ecss_accepted,
	ecss_connected,
	ecss_reqsent,
	ecss_reqrecvd,
};

struct ecs_event {
	struct list_head lnk;
	struct event ev;
	int sfd;
	int state;
	int reqsiz;
	int reqsent;
	int reqrecvd;
	char inbuf[ECS_MAX_REQ_SIZE + sizeof(u_long) + 1];
	char outbuf[ECS_MAX_REQ_SIZE + sizeof(u_long) + 1];
};



static struct list_head conn_head;
static unsigned long ecsreqs;
static volatile int numconns, totconns;




int ecs_socket(int domain, int type, int protocol) {
	int sfd, flags = 1;
 
	if ((sfd = socket(domain, type, protocol)) == -1) {
		perror("socket");
		return -1;
	}
	if (ioctl(sfd, FIONBIO, &flags) &&
	    ((flags = fcntl(sfd, F_GETFL, 0)) < 0 ||
	     fcntl(sfd, F_SETFL, flags | O_NONBLOCK) < 0)) {
		perror("setting NOBLOCK mode");
		close(sfd);
		return -1;
	}
	return sfd;
}


int ecs_server_socket(int port, int lsnsize) {
	int sfd, flags = 1;
	struct linger ling = {0, 0};
	struct sockaddr_in addr;

	if ((sfd = ecs_socket(AF_INET, SOCK_STREAM, 0)) == -1)
		return -1;
 
	setsockopt(sfd, SOL_SOCKET, SO_REUSEADDR, &flags, sizeof(flags));
	setsockopt(sfd, SOL_SOCKET, SO_KEEPALIVE, &flags, sizeof(flags));
	setsockopt(sfd, SOL_SOCKET, SO_LINGER, &ling, sizeof(ling));
 
	addr.sin_family = AF_INET;
	addr.sin_port = htons(port);
	addr.sin_addr.s_addr = htonl(INADDR_ANY);
	if (bind(sfd, (struct sockaddr *) &addr, sizeof(addr)) == -1) {
		perror("bind");
		close(sfd);
		return 4;
	}
	listen(sfd, lsnsize);
	return sfd;
}


struct ecs_event *ecs_newconn(int sfd, void (*func)(int, short, void *),
			      unsigned long flags) {
	struct ecs_event *eev;

	if (!(eev = (struct ecs_event *) malloc(sizeof(struct ecs_event)))) {
		perror("malloc");
		return NULL;
	}

	DBL_LIST_ADDT(&eev->lnk, &conn_head);

	eev->sfd = sfd;
	eev->state = ecss_start;
	event_set(&eev->ev, sfd, flags, func, (void *) eev);
	event_add(&eev->ev, NULL);

	return eev;
}


int ecs_event_change(struct ecs_event *eev, void (*func)(int, short, void *),
		     unsigned long flags) {
	
	event_del(&eev->ev);
	event_set(&eev->ev, eev->sfd, flags, func, (void *) eev);
	event_add(&eev->ev, NULL);

	return 0;
}


int ecs_event_delete(struct ecs_event *eev) {

	event_del(&eev->ev);
	DBL_LIST_DEL(&eev->lnk);
	close(eev->sfd);
	free(eev);

	return 0;
}


unsigned long long ecs_getustime(void) {
	struct timeval tm;

	gettimeofday(&tm, NULL);
	return (unsigned long long) tm.tv_sec * 1000000ULL + (unsigned long long) tm.tv_usec;
}




#if defined(ECS_SERVER)


void server_cb(int fd, short which, void *arg) {
	struct ecs_event *eev = (struct ecs_event *) arg;
	int n;

	switch (eev->state) {
	case ecss_accepted:
		if ((n = read(eev->sfd, eev->inbuf + eev->reqsiz,
			      sizeof(eev->inbuf) - eev->reqsiz)) > 0) {
			eev->reqsiz += n;
			if (eev->reqsiz >= sizeof(u_long) &&
			    eev->reqsiz >= (int) ntohl(*(u_long *) eev->inbuf)) {
				ecs_event_change(eev, server_cb, EV_WRITE | EV_PERSIST);
				eev->state = ecss_reqrecvd;
				eev->reqsent = 0;
			}
		} else {
			if (n == 0 || (errno != EAGAIN && errno != EWOULDBLOCK)) {
				perror("read");
				ecs_event_delete(eev);
			}
		}
		break;

	case ecss_reqrecvd:
		if ((n = write(eev->sfd, eev->inbuf + eev->reqsent,
			       eev->reqsiz - eev->reqsent)) > 0) {
			eev->reqsent += n;
			if (eev->reqsent == eev->reqsiz) {
				ecs_event_delete(eev);
				ecsreqs++;
			}
		} else {
			if (errno != EAGAIN && errno != EWOULDBLOCK) {
				perror("write");
				ecs_event_delete(eev);
			}
		}
		break;
	}
}


void accept_cb(int fd, short which, void *arg) {
	struct ecs_event *eev = (struct ecs_event *) arg, *ceev;
	int sfd, flags = 1;
	socklen_t addrlen;
	struct sockaddr addr;

	addrlen = sizeof(addr);
	while ((sfd = accept(fd, &addr, &addrlen)) >= 0) {
		if (ioctl(sfd, FIONBIO, &flags) &&
		    ((flags = fcntl(sfd, F_GETFL, 0)) < 0 ||
		     fcntl(sfd, F_SETFL, flags | O_NONBLOCK) < 0)) {
			perror("setting NOBLOCK mode");
			close(sfd);
		} else if (!(ceev = ecs_newconn(sfd, server_cb, EV_READ | EV_PERSIST))) {

			close(sfd);
		} else {
			ceev->reqsiz = 0;
			ceev->state = ecss_accepted;
		}
	}

	if (errno != EAGAIN && errno != EWOULDBLOCK) {
		perror("accept");

	}
}


int main (int argc, char **argv) {
	int c, sfd, port = ECS_STD_SERVER_PORT, lsnsize = ECS_STD_LISTEN_SIZE;
	extern char *optarg;

	while ((c = getopt(argc, argv, "p:l:")) != -1) {
		switch (c) {
		case 'p':
			port = atoi(optarg);
			break;
		case 'l':
			lsnsize = atoi(optarg);
			break;
		default:
			fprintf(stderr, "Illegal argument \"%c\"\n", c);
			return 1;
		}
	}

	DBL_INIT_LIST_HEAD(&conn_head);

	event_init();

	if ((sfd = ecs_server_socket(port, lsnsize)) < 0) {

		return 2;
	}

	if (!ecs_newconn(sfd, accept_cb, EV_READ | EV_PERSIST)) {

		close(sfd);
		return 3;
	}

	ecsreqs = 0;
	for (;;) {

		event_loop(EVLOOP_ONCE);

	}

	return 0;
}


#endif /* #if defined(ECS_SERVER) */




#if defined(ECS_CLIENT)


int ecs_generate_str(char *buf, int len) {
	int i, c;

	*(u_long *) buf = htonl(len + sizeof(u_long));
	buf += sizeof(u_long);
	for (i = 0, len--; i < len;) {
		c = rand() & 0x7f;
		if (isalpha(c) || isdigit(c))
			buf[i++] = (char) c;
	}
	buf[i++] = 0;
	return i + sizeof(u_long);
}


void client_cb(int fd, short which, void *arg) {
	struct ecs_event *eev = (struct ecs_event *) arg;
	int n;

	switch (eev->state) {
	case ecss_start:
		numconns++;
		totconns++;
	case ecss_connected:
		if ((n = write(eev->sfd, eev->outbuf + eev->reqsent,
			       eev->reqsiz - eev->reqsent)) > 0) {
			eev->reqsent += n;
			if (eev->reqsent == eev->reqsiz) {
				ecs_event_change(eev, client_cb, EV_READ | EV_PERSIST);
				eev->state = ecss_reqsent;
				eev->reqrecvd = 0;
			}
		} else {
			if (errno != EAGAIN && errno != EWOULDBLOCK) {
				perror("write");
				ecs_event_delete(eev);
				numconns--;
			}
		}
		break;

	case ecss_reqsent:
		if ((n = read(eev->sfd, eev->inbuf + eev->reqrecvd,
			      eev->reqsiz - eev->reqrecvd)) > 0) {
			eev->reqrecvd += n;
			if (eev->reqrecvd == eev->reqsiz) {
				if (!memcmp(eev->inbuf, eev->outbuf, eev->reqsiz))
					ecsreqs++;
				else
					fprintf(stderr, "wrong reply:\n"
						"  sent=%s\n"
						"  recv=%s\n",
						eev->sfd, eev->outbuf + sizeof(u_long),
						eev->inbuf + sizeof(u_long));
				ecs_event_delete(eev);
				numconns--;
			}
		} else {
			if (n == 0 || (errno != EAGAIN && errno != EWOULDBLOCK)) {
				perror("read");
				ecs_event_delete(eev);
				numconns--;
			}
		}
		break;
	}
}


struct ecs_event *ecs_client_conn(struct sockaddr_in *addr, int reqsize) {
	int sfd;
	struct ecs_event *eev;

	if ((sfd = ecs_socket(AF_INET, SOCK_STREAM, 0)) < 0)
		return NULL;

	if (connect(sfd, (struct sockaddr *) addr, sizeof(*addr)) == -1) {
		if (errno != EWOULDBLOCK && errno != EINPROGRESS) {
			perror("connect");
			close(sfd);
			return NULL;
		} else {
			if (!(eev = ecs_newconn(sfd, client_cb,
						EV_WRITE | EV_PERSIST))) {
				close(sfd);
				return NULL;
			} else {
				eev->reqsent = 0;
				eev->reqsiz = ecs_generate_str(eev->outbuf, reqsize);
			}
		}
	} else if ((eev = ecs_newconn(sfd, client_cb,
				      EV_WRITE | EV_PERSIST)) != NULL) {
		eev->state = ecss_connected;
		eev->reqsent = 0;
		eev->reqsiz = ecs_generate_str(eev->outbuf, reqsize);
		numconns++;
		totconns++;
	} else {
		close(sfd);
		return NULL;
	}

	return eev;
}


int main (int argc, char **argv) {
	int c, port = ECS_STD_SERVER_PORT, nconns = ECS_STD_CLIENT_CONNS;
	int ncdiff, maxconns = 0, reqsize = ECS_STD_REQ_SIZE;
	unsigned long reqlast;
	unsigned long long ts, te;
	struct ecs_event *eev;
	char server[128] = "localhost";
	struct hostent *he;
	struct in_addr inadr;
	struct sockaddr_in saddr;
	extern char *optarg;

	while ((c = getopt(argc, argv, "p:s:n:t:r:")) != -1) {
		switch (c) {
		case 'p':
			port = atoi(optarg);
			break;
		case 's':
			strncpy(server, optarg, sizeof(server) - 1);
			break;
		case 'n':
			nconns = atoi(optarg);
			break;
		case 't':
			maxconns = atoi(optarg);
			break;
		case 'r':
			if ((reqsize = atoi(optarg)) > ECS_MAX_REQ_SIZE)
				reqsize = ECS_MAX_REQ_SIZE;
			break;
		default:
			fprintf(stderr, "Illegal argument \"%c\"\n", c);
			return 1;
		}
	}

	if (inet_aton(server, &inadr) == 0) {
		if ((he = gethostbyname(server)) == NULL) {
			perror("gethostbyname");
			return 2;
		}
		memcpy(&inadr.s_addr, he->h_addr_list[0], he->h_length);
	}
	saddr.sin_family = AF_INET;
	saddr.sin_port = htons(port);
	memcpy(&saddr.sin_addr, &inadr.s_addr, 4);

	DBL_INIT_LIST_HEAD(&conn_head);

	event_init();

	srand(time(NULL));

	ts = ecs_getustime();
	reqlast = ecsreqs = 0;
	totconns = numconns = 0;
	for (;;) {

		ncdiff = nconns - numconns;
		if (maxconns && (totconns + ncdiff) > maxconns)
			ncdiff = maxconns - totconns;
		for (; ncdiff > 0; ncdiff--) {

			if (!(eev = ecs_client_conn(&saddr, reqsize)))
				break;

		}

		event_loop(EVLOOP_ONCE);

		te = ecs_getustime();
		if ((te - ts) > ECS_SAMPLE_TIME) {
			fprintf(stdout, "rate=%llu  conns=%d  tconns=%d  reqs=%lu\n",
				(ecsreqs - reqlast) * 1000000ULL / (te - ts), numconns,
				totconns, ecsreqs);

			ts = te;
			reqlast = ecsreqs;
		}

		if (maxconns && totconns >= maxconns && !numconns)
			break;
	}

	te = ecs_getustime();
	if ((te - ts) > 0)
		fprintf(stdout, "rate=%llu  conns=%d  tconns=%d  reqs=%lu\n",
			(ecsreqs - reqlast) * 1000000ULL / (te - ts), numconns,
			totconns, ecsreqs);

	return 0;
}


#endif /* #if defined(ECS_CLIENT) */

